/*
 */

#include "tables.h"

int pat_charfrom(Tcl_Interp *ip, Tcl_Obj *obj, int *val,
		 const char *opts, const char *what) {
  *val= enum1_lookup_cached_func(ip,obj,opts,what);
  if (*val==-1) return TCL_ERROR;
  return TCL_OK;
}

int pat_enum(Tcl_Interp *ip, Tcl_Obj *obj, const void **val,
	     const void *opts, size_t sz, const char *what) {
  *val= enum_lookup_cached_func(ip,obj,opts,sz,what);
  if (!*val) return TCL_ERROR;
  return TCL_OK;
}
  
int pat_obj(Tcl_Interp *ip, Tcl_Obj *obj, Tcl_Obj **val) {
  *val= obj;
  return TCL_OK;
}

int pat_int(Tcl_Interp *ip, Tcl_Obj *obj, int *val) {
  return Tcl_GetIntFromObj(ip, obj, val);
}
  
int pat_long(Tcl_Interp *ip, Tcl_Obj *obj, long *val) {
  return Tcl_GetLongFromObj(ip, obj, val);
}
  
int pat_hbv(Tcl_Interp *ip, Tcl_Obj *var, HBytes_Var *agg) {
  int rc;
  Tcl_Obj *val;

  agg->var= var;

  val= Tcl_ObjGetVar2(ip,var,0,TCL_LEAVE_ERR_MSG);
  if (!val) return TCL_ERROR;

  rc= Tcl_ConvertToType(ip,val,&hbytes_type);
  if (rc) return rc;

  if (Tcl_IsShared(val)) {
    val= Tcl_DuplicateObj(val);
    agg->copied= 1;
  }
  Tcl_InvalidateStringRep(val);
  agg->obj= val;

  agg->hb= OBJ_HBYTES(val);
  return TCL_OK;
}

void fini_hbv(Tcl_Interp *ip, int rc, HBytes_Var *agg) {
  Tcl_Obj *ro;
  
  if (!rc) {
    assert(agg->obj);
    ro= Tcl_ObjSetVar2(ip,agg->var,0,agg->obj,TCL_LEAVE_ERR_MSG);
    if (!ro) rc= TCL_ERROR;
  }
  if (rc && agg->copied)
    Tcl_DecrRefCount(agg->obj);
}

int pat_hb(Tcl_Interp *ip, Tcl_Obj *obj, HBytes_Value *val) {
  int rc;
  rc= Tcl_ConvertToType(ip,obj,&hbytes_type);  if (rc) return rc;
  *val= *OBJ_HBYTES(obj);
  return TCL_OK;
}

Tcl_Obj *ret_hb(Tcl_Interp *ip, HBytes_Value val) {
  Tcl_Obj *obj;
  obj= Tcl_NewObj();
  Tcl_InvalidateStringRep(obj);
  *OBJ_HBYTES(obj)= val;
  obj->typePtr= &hbytes_type;
  return obj;
}

Tcl_Obj *ret_int(Tcl_Interp *ip, int val) {
  return Tcl_NewIntObj(val);
}

Tcl_Obj *ret_long(Tcl_Interp *ip, long val) {
  return Tcl_NewLongObj(val);
}

Tcl_Obj *ret_obj(Tcl_Interp *ip, Tcl_Obj *val) {
  return val;
}

void setstringresult(Tcl_Interp *ip, const char *m) {
  Tcl_ResetResult(ip);
  Tcl_AppendResult(ip, m, (char*)0);
}
